# The usual run-time import of the numpy package/namespace
import numpy as np

# Make Numpy accessible to Cython at compile time: Gives access to
# Numpy C-API.
cimport numpy as cnp
# Typically this import would overload np:
#    cimport numpy as np
# as Cython has no trouble distinguishing between run-time and
# compile-time features. I have used a different name here so that it
# is obvious to the human reader, in the code below, which is which.

# Further boilerplate required to successfully use Numpy C-API
cnp.import_array()


import cython

# Remember decorators? Here's a couple coming up!

# Switch off run-time array bounds checking for the whole function
@cython.boundscheck(False)
# Switch off Python's array index wrapping (negative indices mean
# offset from *end* of array) for the whole function
@cython.wraparound(False)
# Declare the =grid= parameter to be a 2 dimensional Numpy array
# containing complex numbers.
def escape_time(cnp.ndarray[complex, ndim=2] grid, int max_iterations):
        """Don't forget the docstring"""

        # Dimensions of the input (and output) array
        cdef int sizeX
        cdef int sizeY
        sizeX = cnp.PyArray_DIM(grid, 0)
        sizeY = cnp.PyArray_DIM(grid, 1)

        # Create a numpy array for the results. Each point in the
        # input grid will give rise to a single int. The topology of
        # the output data should match that of the input.
        cdef cnp.ndarray[long, ndim=2] out
        grid_shape = (sizeX, sizeY)
        out = np.empty(grid_shape, int)

        # Cython type declarations may only appear at the top level of
        # the function, not inside any further-nested block.
        cdef complex z
        cdef complex c
        cdef int count

        cdef int Nx
        cdef int Ny
        
        for Nx in range(sizeX):
            for Ny in range(sizeY):

                c = grid[Nx,Ny]
                z=0j
                count = 0
                while z.real*z.real+z.imag*z.imag < 4 and count < max_iterations:
                    z = z*z + c
                    count += 1

                out[Nx,Ny] = count

        return out
