import gc

def update_instances(new_class):
    """A decorator that updates all existing instances of a class,
    when that class is redefined.

    NB: Only works for new-style classes."""

    # Get the already-existing class with the same name, or None
    old_class = globals().get(new_class.__name__)

    # If an old version of this class already exists
    if old_class:
        # Find all its instances
        for existing_instance in [ instance for instance in gc.get_objects() 
                                   if type(instance) is old_class ]:
            # And make them instances of the new class
            existing_instance.__class__ = new_class

    # Don't change the new class itself in any way
    return new_class

################################################################################

from time import time

@update_instances
class Mandelbrot(object):

    def __init__(self, resolution=100, x=-0.75, y=0, width=2.5, max_iterations=20):
        # The number of distinct points to draw across the width of the viewport
        self._resolution = resolution
        # The centre of the viewport in the complex plane
        self._x = x
        self._y = y
        # The width of the viewport in the complex plane
        self._width = width
        # If the iteration has not diverged after max_iterations, the
        # point is considered to belong to the set
        self._max_iterations = max_iterations
        # Get a handle on the figure axes, to ensure that updated
        # versions of images replace the old version rather than
        # appearing on a new (sub-)plot.
        fig = figure()
        self._ax = fig.add_subplot(111)
        # Calculate the first image, and make sure it is displayed
        self.calculate_and_draw()

    def update_image(self):
        # To help us appreciate the effects of our optimizations,
        # we'll report how long the calculation of the image takes.
        start = time()

        # Calculate the positions of the boundaries of the viewport.
        w_2 = self._width/2.0
        x_min, x_max = self._x - w_2, self._x + w_2
        y_min, y_max = self._y - w_2, self._y + w_2
        # Imaginary step sizes tell ogrid how many points to generate
        # along a given axis.
        resolution = self._resolution * 1j
        # Calculate the set of points which we want to test
        # NB 1) x,y reversed to get correct grid orientation
        #    2) In Matplotlib coordinates the y axis points down, so
        #       we reverse the y data here
        y,x = ogrid[y_max : y_min : resolution,
                    x_min : x_max : resolution]
        grid = x + (1j * y)
        # Create storage space for the results: one for each point in
        # the grid.
        self._image = empty(grid.shape, float)
        # Calculate the escape time for each point on the grid, and
        # store it in the image
        for (x,y), z in ndenumerate(grid):
            self._image[x,y] = escape_time(z, self._max_iterations)

        # Report the time taken to calculate the image.
        stop = time()
        print "Image recalculation time:%6.2f s" % (stop-start,)

    def draw(self):
        # Draw a heat map of the escape times: the longer the escape
        # time, the redder the pixel
        imshow(self._image, interpolation='nearest', axes=self._ax)
        
    def calculate_and_draw(self):
        self.update_image()
        self.draw()

    def pan(self, direction):
        dx,dy = direction
        self._x += dx * self._width
        self._y += dy * self._width
        print "Panning to", (self._x, self._y)
        self.calculate_and_draw()

def escape_time(c, max_iterations):
    z=0j
    count = 0
    while abs(z) < 2 and count < max_iterations:
        z = z*z + c
        count += 1
    return count

# For easier use with IPython's loadpy and edit: re-use old instance
# if it exists.
if 'mandel' not in dir():
    mandel = Mandelbrot()
