import gc

def update_instances(new_class):
    """A decorator that updates all existing instances of a class,
    when that class is redefined.

    NB: Only works for new-style classes."""

    # Get the already-existing class with the same name, or None
    old_class = globals().get(new_class.__name__)

    # If an old version of this class already exists
    if old_class:
        # Find all its instances
        for existing_instance in [ instance for instance in gc.get_objects()
                                   if type(instance) is old_class ]:
            # And make them instances of the new class
            existing_instance.__class__ = new_class

    # Don't change the new class itself in any way
    return new_class

################################################################################

from time import time

@update_instances
class Mandelbrot(object):

    def __init__(self, resolution=3000, x=-0.75, y=0, width=2.5, max_iterations=20):
        # The number of distinct points to draw across the width of the viewport
        self._resolution = resolution
        # The centre of the viewport in the complex plane
        self._x = x
        self._y = y
        # The width of the viewport in the complex plane
        self._width = width
        # If the iteration has not diverged after max_iterations, the
        # point is considered to belong to the set
        self._max_iterations = max_iterations
        # Create the figure on which the set will be displayed
        fig = figure()
        # Register the handler for mouse clicks in the viewer
        fig.canvas.mpl_connect('button_press_event', self._on_click)
        fig.canvas.mpl_connect(   'key_press_event', self._on_key)
        # Get a handle on the figure axes, to ensure that updated
        # versions of images replace the old version rather than
        # appearing on a new (sub-)plot.
        self._ax = fig.add_subplot(111)
        # Calculate the first image, and make sure it is displayed
        self.calculate_and_draw()
        # Space for remembering consecutively pressed numeric keys,
        # which are to be interpreted as a number requesting a new
        # self._max_iterations
        self._remembered_key_presses = []

    def update_image(self):
        # To help us appreciate the effects of our optimizations,
        # we'll report how long the calculation of the image takes.
        start = time()

        # Calculate the positions of the boundaries of the viewport.
        w_2 = self._width/2.0
        x_min, x_max = self._x - w_2, self._x + w_2
        y_min, y_max = self._y - w_2, self._y + w_2
        # Imaginary step sizes tell ogrid how many points to generate
        # along a given axis.
        resolution = self._resolution * 1j
        # Calculate the set of points which we want to test
        # NB 1) x,y reversed to get correct grid orientation
        #    2) In Matplotlib coordinates the y axis points down, so
        #       we reverse the y data here
        y,x = ogrid[y_max : y_min : resolution,
                    x_min : x_max : resolution]
        grid = x + (1j * y)
        # XXXX
        self._image = escape_time(grid, self._max_iterations)
        # Report the time taken to calculate the image.
        stop = time()
        print "Image recalculation time:%6.2f s" % (stop-start,)

    def draw(self):
        # Draw a heat map of the escape times: the longer the escape
        # time, the redder the pixel
        imshow(self._image, interpolation='nearest', axes=self._ax)

    def calculate_and_draw(self):
        self.update_image()
        self.draw()

    def pan(self, direction):
        dx,dy = direction
        self._x += dx * self._width
        self._y += dy * self._width
        print "Panning to", (self._x, self._y)
        self.calculate_and_draw()

    def zoom(self, factor=2):
        self._width /= factor
        print "Zooming to width", self._width
        self.calculate_and_draw()

    def change_resolution(self, factor=1.1):
        self._resolution = int(factor * self._resolution)
        print "Changing resolution to", self._resolution
        self.calculate_and_draw()

    def max_iterations(self, new_value=None):
        if new_value:
            self._max_iterations = new_value
            print "Setting maximum iterations to", self._max_iterations
            self.calculate_and_draw()
        return self._max_iterations

    def _on_click(self, event):
        # Don't do anything if the click lies outside the axes
        if event.xdata is None or event.ydata is None:
            return
        # Convert axis coordinates into image coordinates and recenter
        # image on the clicked point.
        self._x += self._pixel_to_position(event.xdata, 'x')
        self._y += self._pixel_to_position(event.ydata, 'y')

        # Mouse buttons which zoom after recentering
        if event.button != 1:
            # As you view the fractal boundary in greater detail, you
            # need more iterations to distinguish the features.
            self._max_iterations *= 1.25
            # Not calling self.max_iterations() as this would
            # recalculate the image unnecessarily, *before* we zoom.
            print "Setting maximum iterations to", self._max_iterations
            self.zoom()

        # Any other mouse buttons only recentre
        else:
            print "Recentering on", (self._x, self._y)
            self.calculate_and_draw()

    def _pixel_to_position(self, pixel, axis):
        # In Matplotlib coordinates, the y-axis points downwards, so
        # we reverse the result for the y-axis
        result = (float(pixel) / self._resolution - 0.5) * self._width
        return -result if axis == 'y' else result

    def _on_key(self, event):
        try:
            # The keys which we understand, and their corresponding
            # actions. Numeric keys are not treated here, as multiple
            # consecutive numbers make up a single action (see below).
            choice = {'up'   :(self.pan, ( 0  , 0.2)),
                      'down' :(self.pan, ( 0  ,-0.2)),
                      'left' :(self.pan, (-0.2, 0)),
                      'right':(self.pan, ( 0.2, 0)),
                      "+"    :(self.zoom, 2),
                      "-"    :(self.zoom, 0.5),
                      "R"    :(self.change_resolution, 2),
                      "r"    :(self.change_resolution, 0.5),
                      } [event.key]
            meth = choice[0]
            args = choice[1:]
            meth(*args)
        except KeyError:
            pass

        if event.key == 'enter':
            # If the last few keypresses were all numeric, interpret
            # them as an integer requesting a new max_iterations.
            try:
                self.max_iterations(int(''.join(self._remembered_key_presses)))
            finally:
                pass

        # If the key is a numeric one, remember it along with any
        # subsequent numeric keypresses. A sequence of numeric keys
        # followed by ENTER is interpreted as a request to change
        # max_iterations. Any other non-numeric key cancels the
        # request.
        if event.key is not None and event.key in '0123456789':
            self._remembered_key_presses.append(event.key)
        else:
            # The key pressed was either ENTER (in which case the
            # digts have already been used above in the "if event.key
            # == 'enter'" block), or it was some other non-numeric
            # key, which cancels the numeric request. In either case,
            # we need to clear the store of consecutive key presses.
            self._remembered_key_presses = []


from cython_mandel5 import escape_time

# For easier use with IPython's loadpy and edit: re-use old instance
# if it exists.
if 'mandel' not in dir():
    mandel = Mandelbrot()
